package com.terraformersmc.terraform.command;

import com.mojang.brigadier.builder.LiteralArgumentBuilder;
import com.mojang.brigadier.exceptions.CommandSyntaxException;
import net.fabricmc.fabric.api.registry.CommandRegistry;
import net.minecraft.class_124;
import net.minecraft.class_1657;
import net.minecraft.class_1937;
import net.minecraft.class_1959;
import net.minecraft.class_2168;
import net.minecraft.class_2170;
import net.minecraft.class_2338;
import net.minecraft.class_2378;
import net.minecraft.class_2556;
import net.minecraft.class_2558;
import net.minecraft.class_2561;
import net.minecraft.class_2564;
import net.minecraft.class_2568;
import net.minecraft.class_2583;
import net.minecraft.class_2588;
import net.minecraft.class_2994;
import net.minecraft.class_3532;
import net.minecraft.text.*;
import java.util.Objects;

/**
 * Command to locate the nearest biome to the player that executed it
 *
 * Thanks to Dinolek for porting this https://github.com/Dinolek/LocateBiome
 * Based off https://github.com/Glitchfiend/BiomesOPlenty/blob/4977b0100ca55f96de50337f46ed673512cf503a/src/main/java/biomesoplenty/common/util/biome/BiomeUtils.java
 */
public class FindBiomeCommand {

	private static int timeout = 120_000;

	public static void register() {
		CommandRegistry.INSTANCE.register(false, dispatcher -> {
			LiteralArgumentBuilder<class_2168> builder = class_2170.method_9247("findbiome").requires(source ->
				source.method_9259(2));
			class_2378.field_11153.method_10220().forEach(biome -> builder.then(class_2170.method_9247(Objects.requireNonNull(class_2378.field_11153.method_10221(biome)).toString())
				.executes(context -> execute(context.getSource(), biome))));
			dispatcher.register(builder);
		});
	}

	private static int execute(class_2168 source, class_1959 biome) {
		new Thread(() -> {
			class_2338 executorPos = new class_2338(source.method_9222());
			class_2338 biomePos = null;
			class_2588 biomeName = new class_2588(biome.method_8689());
			try {
				biomePos = spiralOutwardsLookingForBiome(source, source.method_9225(), biome, executorPos.method_10263(), executorPos.method_10260());
			} catch (CommandSyntaxException e) {
				e.printStackTrace();
			}

			if (biomePos == null) {
				source.method_9226(new class_2588(source.method_9211() instanceof class_2994 ? "optimizeWorld.stage.failed" : "commands.terraform.findbiome.fail",
					biomeName, timeout / 1000).method_10862(new class_2583().method_10977(class_124.field_1061)), true);
				return;
			}
			class_2338 finalBiomePos = biomePos;
			source.method_9211().execute(() -> {
				int distance = class_3532.method_15357(getDistance(executorPos.method_10263(), executorPos.method_10260(), finalBiomePos.method_10263(), finalBiomePos.method_10260()));
				class_2561 coordinates = class_2564.method_10885(new class_2588("chat.coordinates", finalBiomePos.method_10263(), "~",
					finalBiomePos.method_10260())).method_10862(new class_2583().method_10977(class_124.field_1060)
					.method_10958(new class_2558(class_2558.class_2559.field_11745, "/tp @s " + finalBiomePos.method_10263() + " ~ " + finalBiomePos.method_10260()))
					.method_10949(new class_2568(class_2568.class_2569.field_11762, new class_2588("chat.coordinates.tooltip"))));
				source.method_9226(new class_2588("commands.locate.success", biomeName, coordinates, distance), true);
			});
		}).start();
		return 0;
	}

	private static class_2338 spiralOutwardsLookingForBiome(class_2168 source, class_1937 world, class_1959 biomeToFind, double startX, double startZ) throws CommandSyntaxException {
		double a = 16 / Math.sqrt(Math.PI);
		double b = 2 * Math.sqrt(Math.PI);
		double x, z;
		double dist = 0;
		long start = System.currentTimeMillis();
		class_2338.class_2340 pos = class_2338.class_2340.method_10109();
		int previous = 0;
		int i = 0;
		for (int n = 0; dist < Integer.MAX_VALUE; ++n) {
			if ((System.currentTimeMillis() - start) > timeout)
				return null;
			double rootN = Math.sqrt(n);
			dist = a * rootN;
			x = startX + (dist * Math.sin(b * rootN));
			z = startZ + (dist * Math.cos(b * rootN));
			pos.method_10102(x, 0, z);
			if (previous == 3)
				previous = 0;
			String dots = (previous == 0 ? "." : previous == 1 ? ".." : "...");
			if (source.method_9228() instanceof class_1657 && !(source.method_9211() instanceof class_2994))
				source.method_9207().method_14254(new class_2588("commands.terraform.findbiome.scanning", dots), class_2556.field_11733);
			if (i == 9216) {
				previous++;
				i = 0;
			}
			i++;
			if (world.method_8310(pos).equals(biomeToFind)) {
				pos.close();
				if (source.method_9228() instanceof class_1657 && !(source.method_9211() instanceof class_2994))
					source.method_9207().method_14254(new class_2588("commands.terraform.findbiome.found", new class_2588(biomeToFind.method_8689()), (System.currentTimeMillis() - start) / 1000), class_2556.field_11733);
				return new class_2338((int) x, 0, (int) z);
			}
		}
		return null;
	}

	private static double getDistance(int posX, int posZ, int biomeX, int biomeZ) {
		return class_3532.method_15368(Math.pow(biomeX - posX, 2) + Math.pow(biomeZ - posZ, 2));
	}

}