package com.terraformersmc.terraform.util;

import com.google.common.base.Preconditions;
import java.util.function.Consumer;
import net.minecraft.class_2338;
import net.minecraft.class_3532;

public class Shapes {
	/**
	 * Iterates over the positions contained with in a circle defined by origin and radius. The circle is two dimensional,
	 * perpendicular to the Y axis.
	 *
	 * @param origin The center block of the circle; this function clobbers the variable, and it must be reset afterwards
	 * @param radius The radius of the circle
	 * @param consumer The target of the positions; it passes the same BlockPos.Mutable object each time
	 */
	public static void circle(class_2338.class_2339 origin, double radius, Consumer<class_2338.class_2339> consumer) {
		int x = origin.method_10263();
		int z = origin.method_10260();

		double radiusSq = radius * radius;
		int radiusCeil = (int) Math.ceil(radius);

		for (int dz = -radiusCeil; dz <= radiusCeil; dz++) {
			int dzSq = dz * dz;

			for (int dx = -radiusCeil; dx <= radiusCeil; dx++) {
				int dxSq = dx * dx;

				if (dzSq + dxSq <= radiusSq) {
					origin.method_10103(x + dx, origin.method_10264(), z + dz);
					consumer.accept(origin);
				}
			}
		}
	}

	/**
	 * Iterates over the positions contained with in a circle defined by origin and radius, but not within the circle with
	 * a radius of two blocks less - effectively, this creates an open canopy shape when stacked. The circle is two
	 * dimensional, perpendicular to the Y axis.
	 *
	 * @param origin The center block of the circle; this function clobbers the variable, and it must be reset afterwards
	 * @param radius The radius of the circle
	 * @param consumer The target of the positions; it passes the same BlockPos.Mutable object each time
	 */
	public static void canopyCircle(class_2338.class_2339 origin, double radius, double innerRadius, Consumer<class_2338.class_2339> consumer) {
		int x = origin.method_10263();
		int z = origin.method_10260();

		int radiusCeil = (int) Math.ceil(radius);
		double maxDistSq = radius * radius;
		double minDistSq = innerRadius * innerRadius;

		for (int dz = -radiusCeil; dz <= radiusCeil; dz++) {
			int dzSq = dz * dz;

			for (int dx = -radiusCeil; dx <= radiusCeil; dx++) {
				int dxSq = dx * dx;
				int distSq = dzSq + dxSq;

				if (distSq <= maxDistSq && distSq >= minDistSq) {
					origin.method_10103(x + dx, origin.method_10264(), z + dz);
					consumer.accept(origin);
				}
			}
		}
	}

	public static final int X = 0;
	public static final int Y = 1;
	public static final int Z = 2;

	public static void line(class_2338.class_2339 origin, int[] offset, Consumer<class_2338.class_2339> consumer) {
		Preconditions.checkArgument(offset.length == 3, "Shape::line: offset must be an array of 3 integers corresponding to the X, Y, and Z offsets");

		int originX = origin.method_10263();
		int originY = origin.method_10264();
		int originZ = origin.method_10260();

		int maxLength = 0;

		for (int i = 0; i < 3; i++) {
			maxLength = Math.max(maxLength, Math.abs(offset[i]));
		}

		if (maxLength == 0) {
			// Avoid dividing by zero
			return;
		}

		double[] position = new double[3];
		double[] velocity = new double[] {
			offset[X] / ((double) maxLength),
			offset[Y] / ((double) maxLength),
			offset[Z] / ((double) maxLength)
		};

		consumer.accept(origin);

		int lastX, lastY, lastZ;
		int currentX, currentY, currentZ;

		for (int i = 0; i < maxLength - 1; i++) {
			position[X] += velocity[X];
			position[Y] += velocity[Y];
			position[Z] += velocity[Z];

			lastX = origin.method_10263();
			lastY = origin.method_10264();
			lastZ = origin.method_10260();

			currentX = originX + class_3532.method_15357(position[X]);
			currentY = originY + class_3532.method_15357(position[Y]);
			currentZ = originZ + class_3532.method_15357(position[Z]);

			boolean cX = lastX != currentX, cY = lastY != currentY, cZ = lastZ != currentZ;

			if(cY && (cX || cZ)) {
				origin.method_10103(lastX, currentY, lastZ);
				consumer.accept(origin);
			}

			if(cX && cZ) {
				if(Math.abs(velocity[X]) > Math.abs(velocity[Z])) {
					origin.method_10103(currentX, currentY, lastZ);
					consumer.accept(origin);
				} else {
					origin.method_10103(lastX, currentY, currentZ);
					consumer.accept(origin);
				}
			}

			origin.method_10103(currentX, currentY, currentZ);
			consumer.accept(origin);
		}
	}
}
